/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Alex Jones - alexjones62@me.com
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#import "SyncController.h"


@implementation SyncController

- (void)syncUsingClientID:(NSString *)clientID forEntityNames:(NSArray *)entityNames {
	
	tfolders = [ NSMutableArray new ];
	tbookmarks = [ NSMutableArray new ];
	
	// load up bookmarks from file
	manager = [ BookmarkManager sharedBookmarkManager ];
	[ manager loadBookmarksLoadingSynchronously: YES ];
	
	// is the sync engine enabled
	if ([[ISyncManager sharedManager] isEnabled] == YES) {
		
		client = [[ISyncManager sharedManager] clientWithIdentifier:clientID ];
		// begin a sync session
		session = [ISyncSession beginSessionWithClient: client entityNames: entityNames beforeDate:[NSDate dateWithTimeIntervalSinceNow:5]];
		
		// start pushing
		[ self startPushForEntities: entityNames ];
		
		[ tfolders writeToFile: @"/folders.plist" atomically: YES ];
		[ tbookmarks writeToFile: @"/bookmarks.plist" atomically: YES ];
		
		if([ session prepareToPullChangesForEntityNames: entityNames beforeDate: [ NSDate distantFuture ]]) {
			// minlging
			[ self startPullForEntities: entityNames ];
		}
		
		// commit changes
		[ session clientCommittedAcceptedChanges ];
		// finish sync session
		[ session finishSyncing ];
		
		[ session release ];
		
	}
	
	[[ NSDistributedNotificationCenter defaultCenter ] postNotificationName: @"bookmarksUpdated" object: @"CaminoSyncTool" ];
	
}

- (void)startPushForEntities:(NSArray *)entities {
	
	FSRef folderRef;
	FSFindFolder(kUserDomain, kApplicationSupportFolderType, kCreateFolder, &folderRef);
	CFURLRef theURL = CFURLCreateFromFSRef(kCFAllocatorDefault, &folderRef);
    NSString* folderPath = [(NSURL*)theURL path];
	CFRelease(theURL);
	
	NSString *profileDir = [ folderPath stringByAppendingPathComponent: @"Camino" ];
	NSString *path = [profileDir stringByAppendingPathComponent:@"changedBookmarkItems.plist"];
	
	NSEnumerator *entityEnumerator = [ entities objectEnumerator ];
	NSString *entityName;
	while (entityName = [ entityEnumerator nextObject ]) {
		
		if ([ session shouldPushChangesForEntityName: entityName ]){
			// pushing changes
			if([ session shouldPushAllRecordsForEntityName: entityName ]) {
				// replace all records on truth
				if([ entityName isEqualToString: @"org.mozilla.camino.bookmarks.Bookmark" ]) {
					// replace all bookmarks
					NSArray *allBookmarks = [[ manager bookmarkRoot ] allChildBookmarks ];
					NSEnumerator *bookmarkEnumerator = [ allBookmarks objectEnumerator ];
					Bookmark *bookmark;
					while(bookmark = [ bookmarkEnumerator nextObject ]) {
						
						NSDictionary *record = [ self getBookmarkDictionary: bookmark ];
						[ tbookmarks addObject: record ];
						NSString *UUID = [ bookmark UUID ];
						[ session pushChangesFromRecord: record withIdentifier: UUID ];
					
					}
				}
				else if([ entityName isEqualToString: @"org.mozilla.camino.bookmarks.BookmarkFolder" ]) {
					// replace all bookmark folders
					NSArray *allBookmarkFolders = [[ manager bookmarkRoot ] allChildBookmarkFolders ];
					NSEnumerator *bookmarkEnumerator = [ allBookmarkFolders objectEnumerator ];
					BookmarkFolder *bookmarkFolder;
					while(bookmarkFolder = [ bookmarkEnumerator nextObject ]) {
						
						if([ bookmarkFolder specialFlag ] != 8) {
							NSDictionary *record = [ self getBookmarkFolderDictionary: bookmarkFolder ];
							[ tfolders addObject: record ];
							NSString *UUID = [ bookmarkFolder UUID ];
							[ session pushChangesFromRecord: record withIdentifier: UUID ];
						}
						
					}
				}					
			}
			else {
				// push only changed records
				if([[ NSFileManager defaultManager ] fileExistsAtPath: path ]) {
					// Get changed bookmark items record IDs
					NSDictionary *dict = [ NSDictionary dictionaryWithContentsOfFile: path ];
					
					if([ entityName isEqualToString: @"org.mozilla.camino.bookmarks.Bookmark" ]) {
						// push changed bookmarks
						
						NSDictionary *changedBookmarks = [ dict objectForKey: @"bookmarks" ];
						
						NSArray *deletedBookmarks = [ changedBookmarks objectForKey: @"deleted" ];
						NSEnumerator *deletedBookmarksEnumerator = [ deletedBookmarks objectEnumerator ];
						NSString *recordID;
						while(recordID = [ deletedBookmarksEnumerator nextObject ]) {
							[ session deleteRecordWithIdentifier: recordID ];
						}
						
						NSArray *addedBookmarks = [ changedBookmarks objectForKey: @"added" ];
						NSEnumerator *addedBookmarksEnumerator = [ addedBookmarks objectEnumerator ];
						while(recordID = [ addedBookmarksEnumerator nextObject ]) {
							NSDictionary *record = [ self getBookmarkDictionary: (Bookmark *)[[ manager bookmarkRoot ] itemWithUUID: recordID ]];
							[ session pushChangesFromRecord: record withIdentifier: recordID ];
						}
						
						NSArray *modifiedBookmarks = [ changedBookmarks objectForKey: @"modified" ];
						NSEnumerator *modifiedBookmarksEnumerator = [ modifiedBookmarks objectEnumerator ];
						while(recordID = [ modifiedBookmarksEnumerator nextObject ]) {
							NSDictionary *record = [ self getBookmarkDictionary: (Bookmark *)[[ manager bookmarkRoot ] itemWithUUID: recordID ]];
							[ session pushChangesFromRecord: record withIdentifier: recordID ];
						}
						
						
					}
					else if([ entityName isEqualToString: @"org.mozilla.camino.bookmarks.BookmarkFolder" ]) {
						// push changed bookmark folders
						
						NSDictionary *changedFolders = [ dict objectForKey: @"folders" ];
						
						NSArray *deletedFolders = [ changedFolders objectForKey: @"deleted" ];
						NSEnumerator *deletedFoldersEnumerator = [ deletedFolders objectEnumerator ];
						NSString *recordID;
						while(recordID = [ deletedFoldersEnumerator nextObject ]) {
							[ session deleteRecordWithIdentifier: recordID ];
						}
						
						NSArray *addedFolders = [ changedFolders objectForKey: @"added" ];
						NSEnumerator *addedFoldersEnumerator = [ addedFolders objectEnumerator ];
						while(recordID = [ addedFoldersEnumerator nextObject ]) {
							BookmarkFolder *bookmarkFolder = (BookmarkFolder *)[[ manager bookmarkRoot ] itemWithUUID: recordID ];
							if([ bookmarkFolder specialFlag ] != 8) {
								NSDictionary *record = [ self getBookmarkFolderDictionary: bookmarkFolder ];
								[ session pushChangesFromRecord: record withIdentifier: recordID ];
							}
						}
						
						NSArray *modifiedFolders = [ changedFolders objectForKey: @"modified" ];
						NSEnumerator *modifiedFoldersEnumerator = [ modifiedFolders objectEnumerator ];
						while(recordID = [ modifiedFoldersEnumerator nextObject ]) {
							BookmarkFolder *bookmarkFolder = (BookmarkFolder *)[[ manager bookmarkRoot ] itemWithUUID: recordID ];
							if([ bookmarkFolder specialFlag ] != 8) {
								NSDictionary *record = [ self getBookmarkFolderDictionary: bookmarkFolder ];
								[ session pushChangesFromRecord: record withIdentifier: recordID ];
							}
						}
						
					}
					
				}
			}
			
		}
		
	}
	
	NSArray *addedBookmarks = [ NSArray array ];
	NSArray *deletedBookmarks = [ NSArray array ];
	NSArray *modifiedBookmarks = [ NSArray array ];
	
	NSArray *addedFolders = [ NSArray array ];
	NSArray *deletedFolders = [ NSArray array ];
	NSArray *modifiedFolders = [ NSArray array ];
	
	NSDictionary *bookmarks = [ NSDictionary dictionaryWithObjects: [ NSArray arrayWithObjects: addedBookmarks, deletedBookmarks, modifiedBookmarks, nil ] forKeys: [ NSArray arrayWithObjects: @"added", @"deleted", @"modified", nil ]];
	NSDictionary *folders = [ NSDictionary dictionaryWithObjects: [ NSArray arrayWithObjects: addedFolders, deletedFolders, modifiedFolders, nil ] forKeys: [ NSArray arrayWithObjects: @"added", @"deleted", @"modified", nil ]];
	
	NSDictionary *newDict = [ NSDictionary dictionaryWithObjects: [ NSArray arrayWithObjects: bookmarks, folders, nil ] forKeys: [ NSArray arrayWithObjects: @"bookmarks", @"folders", nil ]];
	
	[ newDict writeToFile: path atomically: YES ];
	
}

- (NSDictionary *)getBookmarkDictionary:(Bookmark *)bookmark {
	
	NSMutableArray *objects = [ NSMutableArray new ];
	NSMutableArray *keys = [ NSMutableArray new ];
	
	[ objects addObject: @"org.mozilla.camino.bookmarks.Bookmark" ];
	[ keys addObject: ISyncRecordEntityNameKey ];
	
	NSString *title;
	if((title = [ bookmark title ])) {
		[ objects addObject: title ];
		[ keys addObject: @"mTitle" ];
	}		
	
	NSString *description;
	if((description = [ bookmark itemDescription ])) {
		[ objects addObject: description ];
		[ keys addObject: @"mDescription" ];
	}
	
	NSString *keyword;
	if((keyword = [ bookmark shortcut ])) {
		[ objects addObject: keyword ];
		[ keys addObject: @"mShortcut" ];
	}
	
	NSData *icon;
	if((icon = [[ bookmark icon ] TIFFRepresentation ]) == nil)
		icon = [ NSData data ];
	[ objects addObject: icon ];
	[ keys addObject: @"mIcon" ];
	
	NSString *URL;
	if((URL = [ bookmark url ])) {
		[ objects addObject: URL ];
		[ keys addObject: @"mURL" ];
	}
	
	NSDate *lastVisit;
	if((lastVisit = [ bookmark lastVisit ])) {
		[ objects addObject: lastVisit ];
		[ keys addObject: @"mLastVisit" ];
	}
	
	NSNumber *status;
	if((status = [ NSNumber numberWithBool: [ bookmark isSeparator ]])) {
		[ objects addObject: status ];
		[ keys addObject: @"mStatus" ];
	}
	
	NSNumber *numberOfVisits;
	if((numberOfVisits = [ NSNumber numberWithUnsignedInt: [ bookmark numberOfVisits ]])) {
		[ objects addObject: numberOfVisits ];
		[ keys addObject: @"mNumberOfVisits" ];
	}
	
	NSString *faviconURL;
	if((faviconURL = (NSString *)[ bookmark faviconURL ])) {
		[ objects addObject: faviconURL ];
		[ keys addObject: @"mFaviconURL" ];
	}
	
	NSNumber *position;
	position = [ NSNumber numberWithUnsignedInt: [[ bookmark parent ] indexOfObject: bookmark ]];
	[ objects addObject: position ];
	[ keys addObject: @"mPosition" ];
	
	NSArray *parent;
	parent = [ NSArray arrayWithObject: [[ bookmark parent ] UUID ]];
	[ objects addObject: parent ];
	[ keys addObject: @"mParent" ];
	
	NSDictionary *bookmarkRecord = [ NSDictionary dictionaryWithObjects: objects forKeys: keys ];
	
	return bookmarkRecord;
	
}

- (NSDictionary *)getBookmarkFolderDictionary:(BookmarkFolder *)bookmarkFolder {
	
	NSMutableArray *objects = [ NSMutableArray new ];
	NSMutableArray *keys = [ NSMutableArray new ];
	
	[ objects addObject: @"org.mozilla.camino.bookmarks.BookmarkFolder" ];
	[ keys addObject: ISyncRecordEntityNameKey ];
	
	NSString *title;
	if((title = [ bookmarkFolder title ])) {
		[ objects addObject: title ];
		[ keys addObject: @"mTitle" ];
	}
	
	NSString *description;
	if((description = [ bookmarkFolder itemDescription ])) {
		[ objects addObject: description ];
		[ keys addObject: @"mDescription" ];
	}
	
	NSString *keyword;
	if((keyword = [ bookmarkFolder shortcut ])) {
		[ objects addObject: keyword ];
		[ keys addObject: @"mShortcut" ];
	}
	
	NSData *icon;
	if((icon = [[ bookmarkFolder icon ] TIFFRepresentation ]) == nil)
		icon = [ NSData data ];
	[ objects addObject: icon ];
	[ keys addObject: @"mIcon" ];
	
	NSNumber *specialFlag;
	if((specialFlag = [ NSNumber numberWithInt: [ bookmarkFolder specialFlag ]])) {
		[ objects addObject: specialFlag ];
		[ keys addObject: @"mSpecialFlag" ];
	}
	
	NSString *identifier;
	if((identifier = [ bookmarkFolder identifier ])) {;
		[ objects addObject: identifier ];
		[ keys addObject: @"mIdentifier" ];
	}
	
	NSNumber *position;
	if(![[ bookmarkFolder parent ] isRoot ]) {
		position = [ NSNumber numberWithUnsignedInt: [[ bookmarkFolder parent ] indexOfObject: bookmarkFolder ]];
		[ objects addObject: position ];
		[ keys addObject: @"mPosition" ];
	}
	
	NSArray *parent;
	if(![[ bookmarkFolder parent ] isRoot ]) {
		parent = [ NSArray arrayWithObject: [[ bookmarkFolder parent ] UUID ]];
		[ objects addObject: parent ];
		[ keys addObject: @"mParent" ];
	}
	else {
		parent = [ NSArray array ];
		[ objects addObject: parent ];
		[ keys addObject: @"mParent" ];
	}
		
	NSMutableArray *childArray = [ NSMutableArray new ];
	
	NSArray *children = [ bookmarkFolder children ];
	BookmarkItem *child;
	NSEnumerator *childrenEnumerator = [ children objectEnumerator ];
	while(child = [ childrenEnumerator nextObject ]) {
		[ childArray addObject: [ child UUID ]];
	}
	[ objects addObject: childArray ];
	[ keys addObject: @"mChildArray" ];
	
	NSDictionary *bookmarkRecord = [ NSDictionary dictionaryWithObjects: objects forKeys: keys ];
	
	return bookmarkRecord;
	
}

- (void)startPullForEntities:(NSArray *)entities {
	
	NSEnumerator *entityEnumerator = [ entities objectEnumerator ];
	NSMutableArray *filteredEntityNames = [ NSMutableArray array ];
	NSString *entityName;
	while (entityName = [ entityEnumerator nextObject ]){
		
		if ([session shouldPullChangesForEntityName:entityName]) {
			// pulling changes for entityName
			[filteredEntityNames addObject:entityName];
			
			if ([session shouldReplaceAllRecordsOnClientForEntityName:entityName]) {
				// delete all current records on client
				if([ entityName isEqualToString: @"org.mozilla.camino.bookmarks.Bookmark" ] == YES) {
					// delete all current bookmarks
					NSArray *allBookmarks = [[ manager bookmarkRoot ] allChildBookmarks ];
					NSEnumerator *bookmarkEnumerator = [ allBookmarks objectEnumerator ];
					Bookmark *bookmark;
					while(bookmark = [ bookmarkEnumerator nextObject ]) {
						
						[[ manager bookmarkRoot ] deleteChild: bookmark ];
						
					}
					
				}
				else if([ entityName isEqualToString: @"org.mozilla.camino.bookmarks.BookmarkFolder" ] == YES) {
					// delete all current bookmark folders
					NSArray *allBookmarkFolders = [[ manager bookmarkRoot ] allChildBookmarkFolders ];
					NSEnumerator *bookmarkEnumerator = [ allBookmarkFolders objectEnumerator ];
					BookmarkFolder *bookmarkFolder;
					while(bookmarkFolder = [ bookmarkEnumerator nextObject ]) {
						
						if([ bookmarkFolder specialFlag ] != 8) {
							[[ manager bookmarkRoot ] deleteChild: bookmarkFolder ];
						}
						
					}
					
				}
			}
		}
		
	}
	
	NSMutableArray *addedBookmarks = [ NSMutableArray new ];
	NSMutableArray *addedFolders = [ NSMutableArray new ];
	
	NSEnumerator *changeEnumerator = [ session changeEnumeratorForEntityNames: filteredEntityNames ];
	ISyncChange *changeItem;
	while (changeItem = [changeEnumerator nextObject]) {
		
		NSArray *objects;
		NSArray *keys;
		
		switch ([changeItem type]) {
			case ISyncChangeTypeAdd:
				
				objects = [ NSArray arrayWithObjects: [ changeItem record ], [ changeItem recordIdentifier ], nil ];
				keys = [ NSArray arrayWithObjects: @"record", @"identifier", nil ];
				
				if([[[ changeItem record ] objectForKey: ISyncRecordEntityNameKey ] isEqualToString: @"org.mozilla.camino.bookmarks.Bookmark" ]) {
					[ addedBookmarks addObject: [ NSDictionary dictionaryWithObjects: objects forKeys: keys ]];
				}
				else if([[[ changeItem record ] objectForKey: ISyncRecordEntityNameKey ] isEqualToString: @"org.mozilla.camino.bookmarks.BookmarkFolder" ]) {
					[ addedFolders addObject: [ NSDictionary dictionaryWithObjects: objects forKeys: keys ]];
				}
				
				break;
			case ISyncChangeTypeDelete:
				
				if([[[ changeItem record ] objectForKey: ISyncRecordEntityNameKey ] isEqualToString: @"org.mozilla.camino.bookmarks.Bookmark" ]) {
					[ self deleteBookmarkwithIdentifier: [ changeItem recordIdentifier ]];
				}
				else if([[[ changeItem record ] objectForKey: ISyncRecordEntityNameKey ] isEqualToString: @"org.mozilla.camino.bookmarks.BookmarkFolder" ]) {
					[ self deleteBookmarkFolderwithIdentifier: [ changeItem recordIdentifier ]];
				}
				
				break;
			case ISyncChangeTypeModify:
				
				if([[[ changeItem record ] objectForKey: ISyncRecordEntityNameKey ] isEqualToString: @"org.mozilla.camino.bookmarks.Bookmark" ]) {
					[ self setBookmarkRecord: [ changeItem record ] withIdentifier: [ changeItem recordIdentifier ]];
				}
				else if([[[ changeItem record ] objectForKey: ISyncRecordEntityNameKey ] isEqualToString: @"org.mozilla.camino.bookmarks.BookmarkFolder" ]) {
					[ self setBookmarkFolderRecord: [ changeItem record ] withIdentifier: [ changeItem recordIdentifier ]];
					
				}
				
				break;
		}
		
		[ session clientAcceptedChangesForRecordWithIdentifier: [ changeItem recordIdentifier ] formattedRecord:nil newRecordIdentifier:nil ];
	}
			
	NSEnumerator *addedFoldersEnumerator = [ addedFolders objectEnumerator ];
	NSDictionary *addedFolder;
	while(addedFolder = [ addedFoldersEnumerator nextObject ]) {
		
		NSDictionary *record = [ addedFolder objectForKey: @"record" ];
		NSString *identifier = [ addedFolder objectForKey: @"identifier" ];
		
		if([[ record objectForKey: @"mParent" ] count ] == 0) {
			BookmarkFolder *folder = [ self getBookmarkFolderForRecord: record withIdentifier: identifier bookmarkRecords: addedBookmarks bookmarkFolderRecords: addedFolders ];
			[[ manager bookmarkRoot ] appendChild: folder ];
		}
		
	}
	
}

- (Bookmark *)getBookmarkForRecord:(NSDictionary *)record withIdentifier:(NSString *)identifier {
	
	if ([[ record objectForKey: @"mStatus" ] unsignedIntValue] == 1)
		return [ Bookmark separator ];
	
	Bookmark *bookmark = [ Bookmark bookmarkWithTitle:[ record objectForKey: @"mTitle" ] url:[ record objectForKey: @"mURL" ] lastVisit:[ record objectForKey: @"mLastVisit" ]];
	[ bookmark setItemDescription: [ record objectForKey: @"mDescription" ]];
	[ bookmark setShortcut: [ record objectForKey: @"mKeyword" ]];
	[ bookmark setIcon: [[ NSImage alloc ] initWithData: [ record objectForKey: @"mIcon" ]]];
	[ bookmark setUUID: identifier ];
	[ bookmark setNumberOfVisits: [[ record objectForKey: @"mNumberOfVisits" ] unsignedIntValue ]];
	[ bookmark setFaviconURL: [ record objectForKey: @"mFaviconURL" ]];
	
	return bookmark;
	
}

- (BookmarkFolder *)getBookmarkFolderForRecord:(NSDictionary *)record withIdentifier:(NSString *)identifier bookmarkRecords:(NSArray *)bookmarks bookmarkFolderRecords:(NSArray *)bookmarkFolders {
	
	BookmarkFolder *folder = [ BookmarkFolder bookmarkFolderWithTitle: [ record objectForKey: @"mTitle" ]];
	
	[ folder setItemDescription: [ record objectForKey: @"mDescription" ]];
	[ folder setShortcut: [ record objectForKey: @"mKeyword" ]];
	[ folder setIcon: [[ NSImage alloc ] initWithData: [ record objectForKey: @"mIcon" ]]];
	[ folder setSpecialFlag: [[ record objectForKey: @"mSpecialFlag"] unsignedIntValue ]];
	[ folder setUUID: identifier ];
	[ folder setIdentifier: [ record objectForKey: @"mIdentifier" ]];
	
	NSEnumerator *bookmarkFoldersEnumerator = [ bookmarkFolders objectEnumerator ];
	NSDictionary *addedFolder;
	while(addedFolder = [ bookmarkFoldersEnumerator nextObject ]) {
		
		NSDictionary *folderRecord = [ addedFolder objectForKey: @"record" ];
		NSString *folderIdentifier = [ addedFolder objectForKey: @"identifier" ];
		
		if([[ folderRecord objectForKey: @"mChildArray" ] containsObject: folderIdentifier ]) {
			[ folder insertChild: [ self getBookmarkFolderForRecord: record withIdentifier: identifier bookmarkRecords: bookmarks bookmarkFolderRecords: bookmarkFolders ] atIndex: [[ folderRecord objectForKey: @"mPosition" ] unsignedIntValue ] isMove: NO ];
		}
		
	}
	
	NSEnumerator *bookmarksEnumerator = [ bookmarks objectEnumerator ];
	NSDictionary *addedBookmark;
	while(addedBookmark = [ bookmarksEnumerator nextObject ]) {
		
		NSDictionary *bookmarkRecord = [ addedBookmark objectForKey: @"record" ];
		NSString *bookmarkIdentifier = [ addedBookmark objectForKey: @"identifier" ];
		
		if([[ record objectForKey: @"mChildArray" ] containsObject: bookmarkIdentifier ]) {
			[ folder insertChild: [ self getBookmarkForRecord: bookmarkRecord withIdentifier: bookmarkIdentifier ] atIndex: [[ bookmarkRecord objectForKey: @"mPosition" ] unsignedIntValue ] isMove: NO ];
		}
		
	}
	
	return folder;
	
}

- (BOOL)deleteBookmarkwithIdentifier:(NSString *)identifier {
	
	NSArray *allBookmarks = [[ manager bookmarkRoot ] allChildBookmarks ];
	NSEnumerator *bookmarkEnumerator = [ allBookmarks objectEnumerator ];
	Bookmark *bookmark;
	while(bookmark = [ bookmarkEnumerator nextObject ]) {
		
		if([[ bookmark UUID ] isEqualToString: identifier ]) {
			return [[ bookmark parent ] deleteChild: bookmark ];
		}
		
	}
	
	return NO;
}

- (BOOL)deleteBookmarkFolderwithIdentifier:(NSString *)identifier {
	
	NSArray *allBookmarkFolders = [[ manager bookmarkRoot ] allChildBookmarkFolders ];
	NSEnumerator *bookmarkEnumerator = [ allBookmarkFolders objectEnumerator ];
	BookmarkFolder *bookmarkFolder;
	while(bookmarkFolder = [ bookmarkEnumerator nextObject ]) {
		
		if([ bookmarkFolder specialFlag ] != 8) {
			if([[ bookmarkFolder UUID ] isEqualToString: identifier ]) {
				return [[ bookmarkFolder parent ] deleteChild: bookmarkFolder ];
			}
		}
		
	}
	
	return NO;
}

- (void)setBookmarkRecord:(NSDictionary *)record withIdentifier:(NSString *)identifier {
	
	Bookmark *bookmark = (Bookmark *)[[ manager bookmarkRoot ] itemWithUUID: identifier ];
	BookmarkFolder *folder = (BookmarkFolder *)[[ manager bookmarkRoot ] itemWithUUID: [[ record objectForKey: @"mParent" ] objectAtIndex: 0 ]];
	
	if([[ record objectForKey: @"mStatus" ] unsignedIntValue] == 1) {
		[ bookmark setIsSeparator: YES ];
		return;
	}
	
	if([ record objectForKey: @"mTitle" ])
		[ bookmark setTitle: [ record objectForKey: @"mTitle" ]];
	if([ record objectForKey: @"mURL" ])
		[ bookmark setUrl: [ record objectForKey: @"mURL" ]];
	if([ record objectForKey: @"mLastVisit" ])
		[ bookmark setLastVisit: [ record objectForKey: @"mLastVisit" ]];
	if([ record objectForKey: @"mDescription" ])
		[ bookmark setItemDescription: [ record objectForKey: @"mDescription" ]];
	if([ record objectForKey: @"mKeyword" ])
		[ bookmark setShortcut: [ record objectForKey: @"mKeyword" ]];
	if([ record objectForKey: @"mIcon" ])
		[ bookmark setIcon: [[ NSImage alloc ] initWithData: [ record objectForKey: @"mIcon" ]]];
	if([ record objectForKey: @"mNumberOfVisits" ])
		[ bookmark setNumberOfVisits: [[ record objectForKey: @"mNumberOfVisits" ] unsignedIntValue ]];
	if([ record objectForKey: @"mFaviconURL" ])
		[ bookmark setFaviconURL: [ record objectForKey: @"mFaviconURL" ]];
	
	[ bookmark setParent: folder ];
	
	if([[ record objectForKey: @"mPosition" ] unsignedIntValue ] != [ folder indexOfObject: bookmark ])
		[ folder moveChild: bookmark toBookmarkFolder: folder atIndex: [[ record objectForKey: @"mPosition" ] unsignedIntValue ]];
	
}

- (void)setBookmarkFolderRecord:(NSDictionary *)record withIdentifier:(NSString *)identifier {
	
	BookmarkFolder *folder = (BookmarkFolder *)[[ manager bookmarkRoot ] itemWithUUID: identifier ];
	
	if([[ record objectForKey: @"mParent" ] count ] != 0) {
		BookmarkFolder *parentFolder = (BookmarkFolder *)[[ manager bookmarkRoot ] itemWithUUID: [[ record objectForKey: @"mParent" ] objectAtIndex: 0 ]];
		[ folder setParent: parentFolder ];
		if([[ record objectForKey: @"mPosition" ] unsignedIntValue ] != [ parentFolder indexOfObject: folder ])
			[ parentFolder moveChild: folder toBookmarkFolder: parentFolder atIndex: [[ record objectForKey: @"mPosition" ] unsignedIntValue ]];
	}
	
	if([[ record objectForKey: @"mChildArray" ] count ] != 0) {
		NSArray *childArray = [ record objectForKey: @"mChildArray" ];
		NSEnumerator *childArrayEnumerator = [ childArray objectEnumerator ];
		NSString *childIdentifier;
		while(childIdentifier = [ childArrayEnumerator nextObject ]) {
			BookmarkItem *childItem = [[ manager bookmarkRoot ] itemWithUUID: childIdentifier ];
			[ folder appendChild: childItem ];
		}
	}
	
	if([ record objectForKey: @"mTitle" ])
	[ folder setTitle: [ record objectForKey: @"mTitle" ]];
	if([ record objectForKey: @"mDescription" ])
	[ folder setItemDescription: [ record objectForKey: @"mDescription" ]];
	if([ record objectForKey: @"mKeyword" ])
	[ folder setShortcut: [ record objectForKey: @"mKeyword" ]];
	if([ record objectForKey: @"mIcon" ])
	[ folder setIcon: [[ NSImage alloc ] initWithData: [ record objectForKey: @"mIcon" ]]];
	if([ record objectForKey: @"mSpecialFlag"])
	[ folder setSpecialFlag: [[ record objectForKey: @"mSpecialFlag"] unsignedIntValue ]];
	if([ record objectForKey: @"mIdentifier" ])
	[ folder setIdentifier: [ record objectForKey: @"mIdentifier" ]];
	
	
}

@end
